import { AggregateRoot, AggregateID } from '@mobicoop/ddd-library';
import {
  CreateRouteProps,
  Path,
  Role,
  RouteProps,
  PathType,
  Route,
} from './route.types';
import { WaypointProps } from './value-objects/waypoint.value-object';
import { v4 } from 'uuid';
import { RouteNotFoundException } from './route.errors';

export class RouteEntity extends AggregateRoot<RouteProps> {
  protected readonly _id: AggregateID;

  static create = async (create: CreateRouteProps): Promise<RouteEntity> => {
    let routes: Route[];
    try {
      routes = await create.georouter.routes(
        this.getPaths(create.roles, create.waypoints),
        create.georouterSettings,
      );
      if (!routes || routes.length == 0) throw new RouteNotFoundException();
    } catch (e: any) {
      throw e;
    }
    let baseRoute: Route;
    let driverRoute: Route | undefined;
    let passengerRoute: Route | undefined;
    if (routes.some((route: Route) => route.type == PathType.GENERIC)) {
      driverRoute = passengerRoute = routes.find(
        (route: Route) => route.type == PathType.GENERIC,
      );
    } else {
      driverRoute = routes.some((route: Route) => route.type == PathType.DRIVER)
        ? routes.find((route: Route) => route.type == PathType.DRIVER)
        : undefined;
      passengerRoute = routes.some(
        (route: Route) => route.type == PathType.PASSENGER,
      )
        ? routes.find((route: Route) => route.type == PathType.PASSENGER)
        : undefined;
    }
    if (driverRoute) {
      baseRoute = driverRoute;
    } else {
      baseRoute = passengerRoute as Route;
    }
    const routeProps: RouteProps = {
      driverDistance: driverRoute?.distance,
      driverDuration: driverRoute?.duration,
      passengerDistance: passengerRoute?.distance,
      passengerDuration: passengerRoute?.duration,
      fwdAzimuth: baseRoute.fwdAzimuth,
      backAzimuth: baseRoute.backAzimuth,
      distanceAzimuth: baseRoute.distanceAzimuth,
      waypoints: create.waypoints,
      points: baseRoute.points,
    };
    return new RouteEntity({
      id: v4(),
      props: routeProps,
    });
  };

  validate(): void {
    // entity business rules validation to protect it's invariant before saving entity to a database
  }

  private static getPaths = (
    roles: Role[],
    waypoints: WaypointProps[],
  ): Path[] => {
    const paths: Path[] = [];
    if (roles.includes(Role.DRIVER) && roles.includes(Role.PASSENGER)) {
      if (waypoints.length == 2) {
        // 2 points => same route for driver and passenger
        paths.push(this.createGenericPath(waypoints));
      } else {
        paths.push(
          this.createDriverPath(waypoints),
          this.createPassengerPath(waypoints),
        );
      }
    } else if (roles.includes(Role.DRIVER)) {
      paths.push(this.createDriverPath(waypoints));
    } else if (roles.includes(Role.PASSENGER)) {
      paths.push(this.createPassengerPath(waypoints));
    }
    return paths;
  };

  private static createGenericPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(waypoints, PathType.GENERIC);

  private static createDriverPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(waypoints, PathType.DRIVER);

  private static createPassengerPath = (waypoints: WaypointProps[]): Path =>
    this.createPath(
      [waypoints[0], waypoints[waypoints.length - 1]],
      PathType.PASSENGER,
    );

  private static createPath = (
    points: WaypointProps[],
    type: PathType,
  ): Path => ({
    type,
    points,
  });
}
